(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open Def;;
open Pycaml;;
open Parser;;
open Pprinter;;
open State;;
open Kernel;;
open Varmap;;
open String;;
open Mapply;;

(* Mymms terms *)
let pyterm_type = "defterm";;
register_ocamlpill_types [|pyterm_type|];;

let term_type_sample = ((VarMap.add "" (Type, Type) VarMap.empty), Type, Type);;

let (pyterm_from_term, term_from_pyterm) =
  make_ocamlpill_wrapper_unwrapper pyterm_type term_type_sample;;


(* parsing a minima terms *)
let _py_parse_term =
  python_pre_interfaced_function 
    ~doc:"Get a string and parse it into a term."
    [|StringType|]
    (fun py_args ->
      let x = pystring_asstring py_args.(0) in
      let x = concat "" (x :: "." :: []) in
        try 
          let lexbuf = (Lexing.from_string x) in
            match Parser.funclambda Lexer.token lexbuf with
              | None -> raise (Pycaml_exn (Pyerr_Exception, "Cannot Type!!"))
              | Some (def, te, ty) -> pyterm_from_term (def, te, ty)
                  
        with 
            _ -> raise (Pycaml_exn (Pyerr_Exception, "Cannot Parse!!"))
              
    )
in
  register_pre_functions_for_python
  [|("parse_term",_py_parse_term)|]
;;

(* Typing a terms *)
let _py_type_term =
  python_pre_interfaced_function 
    ~doc:"Get the type of a term."
    [|CamlpillType|]
    (fun py_args ->
      let (def, te, ty) = term_from_pyterm py_args.(0) in
      let y = string_of_term ty state.implicite in
        pystring_fromstring y
    )
in
  register_pre_functions_for_python
  [|("type_term",_py_type_term)|]
;;


(* Showing a terms *)
let _py_show_term =
  python_pre_interfaced_function 
    ~doc:"Show the contant of a term."
    [|CamlpillType|]
    (fun py_args ->
      let (def, te) = term_from_pyterm py_args.(0) in
      let y = (
        match te with
          | Obj _ -> show te state.implicite  
          | _ ->
              concat "" (
                (
                  VarMap.fold (
                    
                    fun key data acc ->
                      
                      let (te, ty) = data in
                        concat "" (
                          acc :: "let " :: key :: " : " :: (string_of_term ty state.implicite) :: " := " :: (string_of_term te state.implicite) :: " in\n" :: [])
                          
                          
                  ) def ""
                ) :: "\t" :: (string_of_term te state.implicite) :: []
              )
      ) in
         pystring_fromstring y
    )
in
  register_pre_functions_for_python
  [|("show_term",_py_show_term)|]
;;

(* Typing a terms *)
let _py_apply_term =
  python_pre_interfaced_function 
    ~doc:"Get the type of a term."
    [|CamlpillType; ListType|]
    (fun py_args ->
      let (def, te, ty) = term_from_pyterm py_args.(0) in
      let args = pylist_toarray py_args.(1) in
      let args = Array.map (fun x -> term_from_pyterm x) args in
      match mapply (def, te, ty) (Array.to_list args) state.def state.coercion state.oracles state.overload state.implicite with
        | None -> raise (Pycaml_exn (Pyerr_Exception, "Cannot Type!!"))
        | Some t -> pyterm_from_term t
    )
in
  register_pre_functions_for_python
  [|("apply_term",_py_apply_term)|]
;;


try 
  let filename = Sys.argv.(1) in
  let argv = Array.init ((Array.length Sys.argv) - 1) (fun i -> Sys.argv.(i+1)) in
    set_python_argv argv;
    let _ = python_load filename in
      0
with
  | Invalid_argument _ -> 
      python ()
