(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open List;;
open String;;
open Varset;;
open Varmap;;

class virtual ['a] pObj =
object 
  method virtual get_name: string
  method virtual get_type: 'a
  method virtual pprint: 'a list -> string
  method virtual exec: 'a list -> 'a
  method virtual show: string
end;;

type name = string;;

type term =
    | Type
    | Cste of name
    | Obj of term pObj
    | Var of var
    | Lambda of var * term * term
    | Forall of var * term * term
    | Phi of var * (var * term) list * term * int list * term
    | Ind of var * (var * term) list * term * term list
    | App of term * term list
    | Match of term * (term list) * (term  option)
    | Cons of int * term
;;

type context = (var * term) list
;;



type definition = (term * term) VarMap.t
;;

type subst =
  | Mgu of (var * term) list
  | DnkMgu
  | NoMgu
;;

type 'a return =
  | Ok of 'a
  | Error of string list
;;
