(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open Printf;;
open List;;
open String;;
open Varset;;
open Freshness;;
open Def;;
open Term;;
open Rewrite;;
open Alpha;;
open Beta;;
open Pprinter;;
open Termeq;;
open Listext;;
open Unfold;;
open Varmap;;

let rec fold_atom' te l def =  
match l with
    | [] -> te
    | (hd1, hd2) :: tl ->
        (*printf "( %s, " hd1; print_term hd2; printf ")\n";*)
        fold_atom' (rewrite_term_term_term te hd2 (Cste hd1) VarSet.empty def) tl def
;;

let fold_atom te (def: definition) =
  fold_atom' 
    te 
    (List.fold_left (fun a x -> let (n, (te, ty)) = x in (n,te)::a) [] (vmext def))
    def
;;

(* there may be problems:

   - assure that the def are closed term !
   - do we have to take care about bounded variable for fold_atom ??
*)

let rec fold_term te def =
  (*printf "fold_term: "; print_term te; printf "\n";*)
  let te' = fold_atom te def in
    if (te = te') then (
      match te with
        | Lambda (x, t1, t2) ->
            Lambda (x, (fold_term t1 def), (fold_term t2 def))
        | Forall (x, t1, t2) ->
            Forall (x, (fold_term t1 def), (fold_term t2 def))
        | Phi (x, la, ty, lda, ld) ->
            Phi (x, (fold_args la def), (fold_term ty def), lda, (fold_term ld def))
        | Ind (x, la, ty, lc) ->
            Ind (x, (fold_args la def), (fold_term ty def), (fold_listterm lc def))
        | Match (te, ld, ty) ->
            Match ((fold_term te def), (fold_listterm ld def), 
                   match ty with
                     | None -> None
                     | Some ty -> Some (fold_term ty def)
            )
        | App (t1, t2) ->
            App ((fold_term t1 def), (fold_listterm t2 def))
        | Cons (n, t1) ->
            Cons (n, (fold_term t1 def))
        | _ -> te
    ) else
      te'
and fold_listterm lte def =
  match lte with
    | [] -> []
    | hd :: tl ->
        (fold_term hd def) :: (fold_listterm tl def)
and fold_args la def =
  match la with
    | [] -> []
    | (hd1, hd2) :: tl ->
        (hd1, (fold_term hd2 def)) :: (fold_args tl def)      
;;

let rec fold te def =
  (*printf "fold: "; print_term te; printf "\n";*)
  let te' = fold_term te def in
    if (te = te') then
      te
    else
      (
        fold te' def
      )
;;

let rec fold_ctxt c def =
  match c with
    | [] -> []
    | (hd1, hd2) :: tl ->
        (hd1, fold hd2 def) :: (fold_ctxt tl def)
;;


