(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

(*
  
  Here we should have a function that:
  1) verifies that a term finishes
  2) that the inductive typess are wf
  3) that the types is well checked
      -> what to do if there are some free variables ?? --> add some implicite arguments ?

*)

open Listext;;
open Termchecker;;
open Terminaison;;
open Indwf;;
open Arith;;
open Term;;
open Alpha;;
open Varset;;
open Varmap;;
open Simpl;;
open List;;
open Rewrite;;
open Printf;;
open Pprinter;;

let check_term te ty def coercion oracles overload implicite fvnok =
  let te = alpha_term_vars te ((vmdomain def) +++ (vmdomain overload)) in
  let te = var2cste te ((vmdomain def) +++ (vmdomain overload)) in
  let te = term_diff te VarSet.empty VarSet.empty in
    tcgfv.gfv <- (term_fv te VarSet.empty);
    let ty = (
      match ty with
        | None -> None
        | Some ty ->
            let ty = alpha_term_vars ty ((vmdomain def) +++ (vmdomain overload)) in
            let ty = var2cste ty ((vmdomain def) +++ (vmdomain overload)) in
            let ty = term_diff ty VarSet.empty tcgfv.gfv in
              tcgfv.gfv <- tcgfv.gfv +++ (term_fv ty VarSet.empty); Some (ty)
    ) in
    (*let f = (build_term_equation [] te [] def) in*)

      (* need to redo the terminaison *)
      if ((*not (fm_dp f)*)false) then (
        (* The term does not finishes *)
        printf "%s: %s \n" (string_of_term te implicite) 
          (
            match ty with
              | None -> ""
              | Some ty -> (string_of_term ty implicite)
          );
        None
      ) else (
        (*printf "finishes\n\n"; flush stdout;*)
        if (not (indwf te)) then
          (* The inductive familly are not well formed*)
          None
        else (
          (*printf "well-formed\n\n"; flush stdout;*)
          let l = termcheck [] te ty VarSet.empty def coercion oracles overload implicite in

            List.fold_left (
              
              fun tl' hd ->
                let (ctxt, te, ty, s) = hd in
                  
                let ctxt = subs_filter2 ctxt ((term_fv te VarSet.empty) +++ (term_fv ty VarSet.empty)) in
                  
                  
                  match ctxt with
                    | [] -> Some (te,ty)
                    | _ -> if (fvnok) then tl'
                          else (

                            let ty = largs_foralls ty (rev ctxt) in
                            let te = largs_lambdas te (rev ctxt) in
                              
                              Some (te, ty)

                          )
                        
            ) None l              
              
        )
      )

;;

let simpl_term te def coercion oracles overload implicite =

  match check_term te None def coercion oracles overload implicite false with
    | None -> None
    | Some (te, ty) ->
        Some (simpl2 te def, ty)

;;

let check_term_wctxt ctxt te ty def coercion oracles overload implicite fvnok =
  let bv = (listvarterm2varset ctxt) in
  let te = alpha_term_vars te (bv +++ ((vmdomain def) +++ (vmdomain overload))) in
  let te = var2cste te ((vmdomain def) +++ (vmdomain overload)) in
  let te = term_diff te bv VarSet.empty in
    tcgfv.gfv <- (term_fv te bv);
    let ty = (
      match ty with
        | None -> None
        | Some ty ->
            let ty = alpha_term_vars ty (bv +++ ((vmdomain def) +++ (vmdomain overload))) in
            let ty = var2cste ty ((vmdomain def) +++ (vmdomain overload)) in
            let ty = term_diff ty bv tcgfv.gfv in
              tcgfv.gfv <- tcgfv.gfv +++ (term_fv ty bv); Some (ty)
    ) in
    let f = (build_term_equation ctxt te [] def) in
      if (not (fm_dp f)) then 
        (* The term does not finishes *)
        None
      else (
        (*printf "finishes\n\n"; flush stdout;*)
        if (not (indwf te)) then
          (* The inductive familly are not well formed*)
          None
        else (
          (*printf "well-formed\n\n"; flush stdout;*)
          let l = termcheck ctxt te ty bv def coercion oracles overload implicite in

            List.fold_left (
              
              fun tl' hd ->
                let (ctxt, te, ty, s) = hd in
                  
                let ctxt = subs_filter2 ctxt ((term_fv te bv) +++ (term_fv ty bv)) in
                  
                  
                  match ctxt with
                    | [] -> Some (te,ty)
                    | _ -> if (not fvnok) then tl'
                          else (

                            let ty = largs_foralls ty (rev ctxt) in
                            let te = largs_lambdas te (rev ctxt) in
                              
                              Some (te, ty)

                          )
                        
            ) None l              
              
        )
      )

;;

let simpl_term_wctxt ctxt te def coercion oracles overload implicite =

  match check_term_wctxt ctxt te None def coercion oracles overload implicite false with
    | None -> None
    | Some (te, ty) ->
        Some (simpl2 te def, ty)

;;

let check_mutualterm lname lte lty def coercion oracles overload implicite =
  let bv =  (list2varset lname) in
  let lte = alpha_listterm_vars lte (bv +++ ((vmdomain def) +++ (vmdomain overload))) in
  let lte = var2cste_listterm lte ((vmdomain def) +++ (vmdomain overload)) in
  let lte = listterm_diff lte bv VarSet.empty in
    tcgfv.gfv <- (listterm_fv lte bv);
    let lty =
      List.map (
        
        fun ty ->
          match ty with
            | None -> None
            | Some ty ->
                let ty = alpha_term_vars ty (bv +++ ((vmdomain def) +++ (vmdomain overload))) in
                let ty = var2cste ty ((vmdomain def) +++ (vmdomain overload)) in
                let ty = term_diff ty bv tcgfv.gfv in
                  tcgfv.gfv <- tcgfv.gfv +++ (term_fv ty bv); Some (ty)
      ) lty in

    let l = termcheckmutual [] 0 lname lte lty bv def coercion oracles overload implicite in

      
      List.fold_left (
        
        fun tl' hd ->
          let (ctxt, lte, lty, s) = hd in

          let ctxt = subs_filter2 ctxt ((listterm_fv lte bv) +++ (listterm_fv lty bv)) in
            match ctxt with
              | [] -> Some (List.combine lname (List.combine lte lty))
              | _ -> 
                  (*
                    printf "**************************\n%s |-\n%s\n\n%s\n\n**************************\n\n"
                    (string_of_listvarterm ctxt implicite)
                    (string_of_listterm lte implicite)
                    (string_of_listterm lty implicite)
                    ; 
                  *)                  
                  tl'
                  
      ) None l              
        
        
;;
  
