(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open Printf;;
open List;;
open String;;
open Varset;;
open Varmap;;
open Freshness;;
open Def;;
open Term;;
open Rewrite;;
open Alpha;;
open Beta;;
open Pprinter;;
open Termeq;;
open Unification;;
open Listext;;
open Arith;;
open Unfold;;
open Unfoldbetared;;
open Fold;;
open Termchecker;;

let rec match_var_args x from into =
  match (from, into) with
    | ((hd11, hd12) :: tl1, (hd21, hd22) :: tl2) ->
        if (x = hd11) then hd21 else
          (match_var_args x tl1 tl2)
    | _ -> x;;

let rec match_comp_args s from into =
  match s with
    | [] -> []
    | (hd1, hd2) :: tl ->
        (match_var_args hd1 from into, hd2) :: (match_comp_args tl from into);;

let rec destruct_cases' n te ctxt args ldes lcons ty ind_name ind_args ind_ty ind_lcons def =
  match lcons with
    | [] -> (
        match ldes with
          | [] -> Some ([], [], [])
          | _ -> (*printf "None1\n";*) None
      )
    | hdc :: tlc -> 
        let ind = Ind (ind_name, ind_args, ind_ty, ind_lcons) in
        let hdc2 = (alpha_term_vars (largs_foralls (rewrite_term_var_term hdc ind_name ind) ind_args) ((listvarterm2varset ctxt) +++ (listterm_fv args VarSet.empty))) in
        let (fcv, con) = (decomp_foralls hdc2) in 
        let (i, iargs) = (decomp_app con) in 
          match (unification_listterm 
                    iargs
                    args
                    VarSet.empty
                    def) with
            | NoMgu -> 
                (*printf "NoMgu\n";*)
                destruct_cases' (n + 1) te ctxt args ldes tlc ty ind_name ind_args ind_ty ind_lcons def
            | DnkMgu -> (                
                (*
                  printf "DnkMgu(\n";
                  print_listterm iargs; printf "  & \n";
                  print_listterm args; printf ")\n";
                *)
                match ldes with
                  | [] -> (*printf "None2\n";*) None
                  | hdd :: tld ->
                      let hdd2 = (alpha_term_vars hdd (listvarterm2varset (fcv @ ctxt))) in
                      let (lvt, _) = (decomp_nlambdas hdd2 (List.length fcv)) in 
                      let s' = comp_subs (
                        
                        match (unification_listterm (list_name2list_var (list_proj1 fcv)) (list_name2list_var (list_proj1 lvt)) VarSet.empty def) with
                          | Mgu s2 -> s2
                          | _ -> []
                      ) in
                      
                      let p = (app_nf (App (Cons (n,ind), list_name2list_var (list_proj1 lvt)))) in
                      let ctxt' = (rev(lvt) @ ctxt) in
                      let hdd' = (rewrite_term_term_term hdd2 te p VarSet.empty def) in
                      let (_, hdd'') = (decomp_nlambdas hdd' (List.length fcv)) in 
                      let ctxt'' = (rewrite_context_term_term ctxt' te p VarSet.empty def) in
                      let ty' = (rewrite_term_term_term ty te p VarSet.empty def) in
                      let sf = s' in
                      let ctxt''' = (apply_subs_context ctxt'' sf VarSet.empty) in
                      let hdd''' = (apply_subs_term hdd'' sf VarSet.empty) in                        
                      let ty'' = (apply_subs_term ty' sf VarSet.empty) in              
                      let lvt' = (
                        (rev (apply_subs_context (rev lvt) sf VarSet.empty))
                      ) in 
                      let fcv' = (
                        (*(rev (apply_subs_context (rev fcv) sf VarSet.empty))*) fcv
                      ) in
                        match (destruct_cases' (n+1) te ctxt args tld tlc ty ind_name ind_args ind_ty ind_lcons def) with
                          | None -> None
                          | Some (l1, l2, l3) ->
                              Some (
                                (ctxt''', hdd''', ty'') :: l1,
                                (apply_subs_context ctxt sf VarSet.empty, largs_lambdas Type lvt', largs_foralls Type fcv') :: l2,
                                (lvt', fcv') :: l3
                              )
              )
            | Mgu u -> (
                match ldes with
                  | [] -> (*printf "None3\n";*) None
                  | hdd :: tld ->
                      let hdd2 = (alpha_term_vars hdd (listvarterm2varset (fcv @ ctxt))) in
                      let (lvt, _) = (decomp_nlambdas hdd2 (List.length fcv)) in 
                      let s' = comp_subs (
                        
                        match (unification_listterm (list_name2list_var (list_proj1 fcv)) (list_name2list_var (list_proj1 lvt)) VarSet.empty def) with
                          | Mgu s2 -> s2
                          | _ -> []
                      ) in

                      let p = (app_nf (App (Cons (n,ind), list_name2list_var (list_proj1 lvt)))) in
                      let u' = comp_subs (apply_subs_context (match_comp_args u fcv lvt) s' VarSet.empty) in       
                      let ctxt' = (rev(lvt) @ ctxt) in
                      let hdd' = (rewrite_term_term_term hdd2 te p VarSet.empty def) in
                      let (_, hdd'') = (decomp_nlambdas hdd' (List.length fcv)) in 
                      let ctxt'' = (rewrite_context_term_term ctxt' te p VarSet.empty def) in
                      let ty' = (rewrite_term_term_term ty te p VarSet.empty def) in
                      let sf = u' in
                      let ctxt''' = (apply_subs_context ctxt'' sf VarSet.empty) in
                      let hdd''' = (apply_subs_term hdd'' sf VarSet.empty) in                        
                      let ty'' = (apply_subs_term ty' sf VarSet.empty) in              
                      let lvt' = (
                        (rev (apply_subs_context (rev lvt) sf VarSet.empty))
                      ) in 
                      let fcv' = (
                        (rev (apply_subs_context (rev fcv) u VarSet.empty)) (* sf ou u ??? pareil ??? aparement non, une autre enigme en vue ... *)
                      ) in
                        match (destruct_cases' (n+1) te ctxt args tld tlc ty ind_name ind_args ind_ty ind_lcons def) with
                          | None -> None
                          | Some (l1, l2, l3) ->
                              Some (
                                (ctxt''', hdd''', ty'') :: l1,
                                (apply_subs_context ctxt sf VarSet.empty, largs_lambdas Type lvt', largs_foralls Type fcv') :: l2,
                                (lvt', fcv') :: l3
                              )
              )
;;

let destruct_cases te ctxt args ldes ty ind_name ind_args ind_ty ind_lcons def =
  (*
  printf "(%d, %d)\n" (List.length ldes) (List.length ind_lcons);
  print_listterm ldes; printf "\n"; print_listterm ind_lcons; printf "\n"; 
  *)
  match destruct_cases' 0 te ctxt args ldes ind_lcons ty ind_name ind_args ind_ty ind_lcons def with
    | a -> a
;;

let rec term_size t =
    match t with
      | Cste c -> Some (Ncons 1)
      | Var v -> 
          Some (Nvar v)
      | App (t1, t2) -> (
          match (term_size t1, listterm_size t2) with
            | (Some s1, Some s2) -> Some (Nplus (s1, s2))
            | _ -> None
        )
      | Cons (n, t1) ->
          Some (Ncons 1)
            
      | _ -> None
and listterm_size lt =
  match lt with
    | [] -> Some (Ncons 0)
    | hd :: tl ->
        match (term_size hd, listterm_size tl) with
          | (Some s1, Some s2) -> Some (Nplus (s1, s2))
          | _ -> None
;;

let rec get_args args ind = 
  match ind with
    | [] -> []
    | hd :: tl ->
        try 
          (List.nth args hd) :: (get_args args tl)
        with
            _ -> printf "get_args\n"; []
;;

let rec sum_size_list_var l =
  match l with
    | [] -> Ncons 0
    | hd :: tl ->
        Nplus (Nvar hd, sum_size_list_var tl)
;;

let rec list_rec_varset l =
  match l with
    | [] -> VarSet.empty
    | (rec_name, rec_dec_args, dec_args_indexes) :: tl ->
        rec_name ++ (list_rec_varset tl)
;;

let rec build_app_term fct args l def =
  match l with
    | [] -> None
    | (rec_name, rec_dec_args, dec_args_indexes) :: tl ->
        if (Var rec_name = fct) then (
          match (listterm_size rec_dec_args , listterm_size (get_args args dec_args_indexes)) with
            | (Some e1, Some e2) ->
                Some (Bgt (e1, e2))
            | _ -> Some BFalse
        ) else (
          build_app_term fct args tl def
        )
;;
        
let rec build_phiargs_equation la args lda =
  match (lda) with
    | [] -> BTrue
    | hd :: tl ->
        try (
          match (term_size (List.nth args hd)) with
            | None -> BTrue
            | Some e -> 
                try (
                  Band (
                    Beq ((Nvar (List.nth la hd)), e),
                    build_phiargs_equation la args tl
                  )
                ) with
                    _ -> BTrue
        ) with
            _ -> 
              BTrue
;;

let rec list_var_pos l =
  match l with
    | [] -> BTrue
    | hd :: tl ->
        Band 
          (
            Bgt (Nvar hd, Ncons 0),
            list_var_pos tl
          )
;;
          
  

(* 

   build an equation that asserts the terminaison of the term execution
   
   the term should be alpha reduced

   it takes at hypothesis that all the cons do terminate

*)

let rec build_term_equation ctxt te l def =
  let y = (
  match te with
    | Lambda (x, t1, t2) ->            
        Band (
          build_term_equation ctxt t1 l def,
          build_term_equation ((x, t1)::ctxt) t2 l def
        )
    | Forall (x, t1, t2) ->            
        Band (
          build_term_equation ctxt t1 l def,
          build_term_equation ((x, t1)::ctxt) t2 l def
        )
    | App (t1, t2) -> (
        let (fct, args) = (decomp_app (unfold_beta_red te def)) in
          match (build_app_term fct args l def) with
            | None -> (
                match fct with
                  | Phi (x, la, ty, lda, te) ->
                      Band (
                        build_listterm_equation ctxt t2 l def,
                        Band (
                          Band (
                            build_listargs_equation ctxt la l def,
                            build_term_equation ((rev la) @ ctxt) ty l def
                          ),
                          bimpl 
                            (build_phiargs_equation (list_proj1 la) args lda)
                            (build_term_equation ((rev la) @ ctxt) te ((x, get_args (list_name2list_var (list_proj1 la)) lda, lda) :: l) def)
                          
                        )
                      )
                  | _ -> 
                      Band (
                        build_term_equation ctxt t1 l def,
                        build_listterm_equation ctxt t2 l def
                      )     
              )     
            | Some f -> 
                Band (
                  build_listterm_equation ctxt t2 l def,
                  f
                )
      )
    | Match (te, ldes, ty) -> (
        (* should be redo to match the new termchecking algo *)
        match (termcheck ctxt te None VarSet.empty def [] [] VarMap.empty VarMap.empty) with
          | (_, _, tye, _)::tl ->
              let (fct, args) = decomp_app (unfold_beta_red tye def) in (
                match fct with
                  | (Ind (x, la, ty, lc)) -> (
                      match (destruct_cases te ctxt args ldes Type x la ty lc def) with
                        | None -> BFalse
                        | Some (l1, l2, l3) -> (
                            match (term_size te) with
                              | None -> BFalse
                              | Some e1 -> 
                                  (*print_term te; printf "\n\n";*)
                                  Band (
                                    build_matchterm_equation l1 l3 e1 l def,
                                    build_term_equation ctxt te l def
                                  )
                          )
                    )
                  | _ -> BFalse
              )
          | _ -> BFalse
      )
    | Phi (x, la, ty, lda, te) ->
        Band (
          Band (
            build_listargs_equation ctxt la l def,
            build_term_equation ((rev la) @ ctxt) ty l def
          ),
          build_term_equation ((rev la) @ ctxt) te ((x, get_args (list_name2list_var (list_proj1 la)) lda, lda) :: l) def
        )
    | Ind (x, la, ty, lcons) ->
        Band (
          Band (
            build_listargs_equation ctxt la l def,
            build_term_equation ((rev la) @ ctxt) ty l def
          ),
          build_listterm_equation ((rev la) @ ctxt) lcons l def
        )
    | Var v -> BTrue
    | Type -> BTrue
    | Cons (n, t1) -> BTrue
    | Cste x -> BTrue
    | Obj o -> BTrue
  ) in
    (*
      printf "\n----------------------\n";
      print_term (fold te def);
      printf ": \n";
      print_bexpr y;
      printf "\n";
      printf "\n----------------------\n";
    *)
    y
and build_listterm_equation ctxt lte l def =
  match lte with
    | [] -> BTrue
    | hd :: tl ->
        Band (
          build_term_equation ctxt hd l def,
          build_listterm_equation ctxt tl l def
        )
and build_matchterm_equation l1 l2 e l def =
  match (l1, l2) with
    | ([], []) ->
        BTrue
    | ((ctxt, te, ty) :: tl1, (lvt, fcv) :: tl2) -> (
        Band (
          bimpl 
            (Band 
                (
                  Beq (e, Nplus (Ncons 1, sum_size_list_var (list_proj1 lvt))),
                  list_var_pos (list_proj1 lvt)
                )
            )
            (build_term_equation ctxt te l def)
          ,
          build_matchterm_equation tl1 tl2 e l def
        )
      )
    | _ -> BFalse
and build_listargs_equation ctxt la l def =
  match la with
    | [] -> BTrue
    | (hd1, hd2) :: tl ->
        Band (
          build_term_equation ctxt hd2 l def,
          build_listargs_equation ((hd1, hd2) :: ctxt) tl l def
        )
;;

