(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open Printf;;
open List;;
open String;;
open Varset;;
open Freshness;;
open Def;;
open Term;;
open Rewrite;;
open Pprinter;;
open Varmap;;


let rec unfold_cste c def =
  try (
    let (te, ty) = VarMap.find c def in
      te
  ) with
    | _ -> Cste c
;;


(*

  unfold:
  
  - is use to unfold type in typechecking (that is already have been typed to type)
  - use beta reduction

*)

let rec unfold_term te def =
(*
  printf "unfold_term: ";
  print_term te VarMap.empty;
  printf "\n\n";
*)
  app_nf (
    match te with
      | Obj o -> Obj o
      | Cste v -> (
          try (
            let (te, ty) = VarMap.find v def in
              (*unfold_term te def*)
              te
          ) with
            | _ -> Cste v
        )
      | App (t1, t2) -> (
          match (unfold_term t1 def) with
            | Lambda (x, t11, t12) -> 
                (App (Lambda (x, t11, t12), t2))
            | Cste v ->
                if (unfold_term t1 def = t1) then
                  te
                else
                  (App ((unfold_term t1 def), t2))
            | _ -> App (unfold_term t1 def, unfold_listterm t2 def)
        )
      | Match (te', lde, ty) ->
          let te1 = (Match (unfold_term te' def, unfold_listterm lde def, 
                           match ty with
                             | None -> None
                             | Some ty -> Some (unfold_term ty def)
          )
          ) in
            te1
      | _ -> te
  )
and unfold_listterm lt def =
  match lt with
    | [] ->[]
    | hd :: tl ->
        (unfold_term hd def) :: (unfold_listterm tl def)
;;
        
