(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open Varset;;
open Varmap;;
open Def;;
open Simpl;;
open Term;;
open Kernel;;

(* These functions should be moved somewhere else *)
let rec closure_def_names def la =
  let la' =
    VarSet.fold (
      
      fun x acc ->
        try (
          let (te, ty) = VarMap.find x def in
          let tec = term_cons te in
          let tyc = term_cons ty in
            acc +++ tec +++ tyc
        ) 
        with
          | _ -> acc

    ) la la in
    if VarSet.equal la la' then
      la'
    else
      closure_def_names def la'
;;

let restrict_def def name =
  VarSet.fold (

    fun x acc ->
      try (
        let (te, ty) = VarMap.find x def in
          VarMap.add x (te, ty) acc
      ) 
      with
        | _ -> acc

  ) name VarMap.empty 
;;

(* application for minima "terms" *)
let mapply te la def coercion oracles overload implicite =
  let (defa, la) = 
    List.fold_left (

      fun acc hd ->
        let (def, te, _) = hd in
        let (accdef, accargs) = acc in
          (varmap_union def accdef, accargs @ te::[])

    ) (VarMap.empty, []) la 
  in
  let (defe, te, _) = te in
  let defe = varmap_union defe defa in
  let def = varmap_union defe def in
  let te = simpl (App (te, la)) def in
    match check_term te None def coercion oracles overload implicite true with
      | None -> None 
      | Some (te,ty) -> 
          Some (
            restrict_def defe (closure_def_names defe ((term_cons te) +++ (term_cons ty))), 
            te, 
            ty
          )
          
;;

    
