(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

open Def;;
open Pprinter;;
open State;;
open String;;
open Simpl;;
open Varmap;;
open Num;;

(* The class for the num type *)

class numTypeObj =
object (self)
  inherit [term] pObj
  method get_name = "Num"
  method get_type = Type
  method pprint = (fun args -> self#get_name)
  method my_self = (self :> term pObj)
  method exec = (fun args -> (Obj (self#my_self)))
  method show = self#get_name
end;;

let numType = 
  let o = new numTypeObj in
    Obj o#my_self
;;

(* The class for the num term *)
class numTermObj n =
object (self)
  inherit [term] pObj
  val mutable numvalue = n
  method get_numvalue = numvalue
  method get_name = string_of_num numvalue
  method get_type = numType
  method pprint = (fun args -> self#get_name)
  method my_self = (self :> term pObj)
  method exec = (fun args -> (Obj (self#my_self)))
  method show = self#get_name
end;;

let create_numTerm n =
  let o = new numTermObj n in
    Obj o#my_self
;;

(* The classes for the basic functions on num *)
class numplusTermObj  =
object (self)
  inherit [term] pObj
  method get_name = "numplus"
  method get_type = Forall ("H0", numType, Forall ("H0", numType, numType))
  method pprint = (fun args -> 
    match args with
      | hd1::hd2::[] ->
          concat "" ((string_of_term hd1 state.implicite) :: " + " :: (string_of_term hd2 state.implicite) :: [])

      | _ -> concat "" ("numplus " :: (string_of_listterm args state.implicite) :: [])
  )
  method my_self = (self :> term pObj)
  method exec = (fun args -> 
    
    match (simpl_listterm args state.def) with
      | (Obj o1) :: (Obj o2) :: [] -> (

          let o1 = ((Obj.magic o1) :> numTermObj) in
          let o2 = ((Obj.magic o2) :> numTermObj) in
            create_numTerm (o1#get_numvalue +/ o2#get_numvalue)
        )       
      | a -> App (Obj self#my_self, a)


  )
  method show = self#get_name
end;;

let numplusTerm =
  let o = new numplusTermObj in
    Obj o#my_self
;;

class numminusTermObj  =
object (self)
  inherit [term] pObj
  method get_name = "numminus"
  method get_type = Forall ("H0", numType, Forall ("H0", numType, numType))
  method pprint = (fun args -> 
    match args with
      | hd1::hd2::[] ->
          concat "" ((string_of_term hd1 state.implicite) :: " - " :: (string_of_term hd2 state.implicite) :: [])

      | _ -> concat "" ("numminus " :: (string_of_listterm args state.implicite) :: [])
  )
  method my_self = (self :> term pObj)
  method exec = (fun args -> 
    
    match (simpl_listterm args state.def) with
      | (Obj o1) :: (Obj o2) :: [] -> (

          let o1 = ((Obj.magic o1) :> numTermObj) in
          let o2 = ((Obj.magic o2) :> numTermObj) in
            create_numTerm (o1#get_numvalue -/ o2#get_numvalue)
        )       
      | a -> App (Obj self#my_self, a)


  )
  method show = self#get_name
end;;

let numminusTerm =
  let o = new numminusTermObj in
    Obj o#my_self
;;

class nummultTermObj  =
object (self)
  inherit [term] pObj
  method get_name = "nummult"
  method get_type = Forall ("H0", numType, Forall ("H0", numType, numType))
  method pprint = (fun args -> 
    match args with
      | hd1::hd2::[] ->
          concat "" ((string_of_term hd1 state.implicite) :: " * " :: (string_of_term hd2 state.implicite) :: [])

      | _ -> concat "" ("nummult " :: (string_of_listterm args state.implicite) :: [])
  )
  method my_self = (self :> term pObj)
  method exec = (fun args -> 
    
    match (simpl_listterm args state.def) with
      | (Obj o1) :: (Obj o2) :: [] -> (

          let o1 = ((Obj.magic o1) :> numTermObj) in
          let o2 = ((Obj.magic o2) :> numTermObj) in
            create_numTerm (o1#get_numvalue */ o2#get_numvalue)
        )       
      | a -> App (Obj self#my_self, a)


  )
  method show = self#get_name
end;;

let nummultTerm =
  let o = new nummultTermObj in
    Obj o#my_self
;;

class numdivTermObj  =
object (self)
  inherit [term] pObj
  method get_name = "numdiv"
  method get_type = Forall ("H0", numType, Forall ("H0", numType, numType))
  method pprint = (fun args -> 
    match args with
      | hd1::hd2::[] ->
          concat "" ((string_of_term hd1 state.implicite) :: " / " :: (string_of_term hd2 state.implicite) :: [])

      | _ -> concat "" ("numdiv " :: (string_of_listterm args state.implicite) :: [])
  )
  method my_self = (self :> term pObj)
  method exec = (fun args -> 
    
    match (simpl_listterm args state.def) with
      | (Obj o1) :: (Obj o2) :: [] -> (

          let o1 = ((Obj.magic o1) :> numTermObj) in
          let o2 = ((Obj.magic o2) :> numTermObj) in
            create_numTerm (o1#get_numvalue // o2#get_numvalue)
        )       
      | a -> App (Obj self#my_self, a)


  )
  method show = self#get_name
end;;

let numdivTerm =
  let o = new numdivTermObj in
    Obj o#my_self
;;



let mnum_init =
  add_overload "plus" numplusTerm 0;
  add_overload "minus" numminusTerm 0;
  add_overload "mult" nummultTerm 0;
  add_overload "div" numdivTerm 0;
;;
