%{

(*
  This file is part of Mymms.

  Mymms is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Mymms is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Mymms.  If not, see <http://www.gnu.org/licenses/>.

  Copyright (C) 2008 Nicolas Marti
*)

  open Printf;;
  open Lexing;;
  open List;;
  open String;;
  
  open Def;;
  open Pprinter;;
  open Freshness;;
  open Term;;
  open Alpha;;
  open Beta;;
  open Termeq;;
  open Varset;;
  open Varmap;;
  open Arith;;
  open Terminaison;;
  open Rewrite;;
  open Fold;;
  open Unfold;;
  open Simpl;;
  open Listext;;
  open Termchecker;;
  open Kernel;;
  open Listext;;

  open State;;

  open Mbool;;
  open Mnum;;
  open Num;;

  (* *)
  
  type pcmd =
      | Definition of (string) list
      | Command
      | Coercion
      | Overload of (string * term)
      | Implicite of string
  ;;




%}

/*  */
%token NEWLINE EOF

/* The indentifiers variables */
%token <string> NAME

/* Integer (for indexes of contructors) */
%token <string> NUM

%token UNDERSCORE AROBAS

/* quantifier */
%token LAMBDA IND PHI FORALL
  
/* term */
%token MATCH RETURN WITH CONS OF TYPE IMPL LET IN

%right IMPL

/* parenthese */
%token LPAREN RPAREN LBRACK RBRACK LACCOL RACCOL

/* ponctuation */
%token PAR SCOMMA COMMA DOT DDOT TURNSTIL EQ 
  
%left SCOMMA
%right PAR

/* command */ 
%token ALPHA TYPECHECK QUIT SIMPL SHOWDEF BETA SHOWCOERCION UNDO

/* definition */ 
%token INDUCTIVE RECURSIVE DEFINITION COERCION OVERLOAD IMPLICITE

/* Overloaded function */
%token EQ PLUS MINUS MULT DIV

/* Boolean */
%token <bool> BOOL
%token BOOLT BNEG BAND BOR


%start input command funclambda
%type <unit> input command
%type <(Def.definition * Def.term * Def.term) option> funclambda
  
%% /* Grammar rules and actions follow */

  input: 
    
| command input {}

| funclambda input {

    match $1 with
      | None -> printf "Cannot TypeCheck!!!\n"
      | Some (def, te, ty) ->
          
          VarMap.fold (

            fun key data acc ->

              let (te, ty) = data in
                printf "let %s: %s := %s in\n"
                  key
                  (string_of_term ty state.implicite)
                  (string_of_term te state.implicite);
                  
                  
          ) def ();
          printf "\t%s\n\n" (string_of_term te state.implicite);

}

| EOF {}
          
    
  ;               
  
  command :

| TYPECHECK term DOT {

    match check_term $2 None state.def state.coercion state.oracles state.overload state.implicite false with
      | None -> printf "Cannot Type Check\n\n";
      | Some (te, ty) ->
          printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);


  }

| TYPECHECK term DDOT term DOT {

    match check_term $2 (Some $4) state.def state.coercion state.oracles state.overload state.implicite false with
      | None -> printf "Cannot Type Check\n\n";
      | Some (te, ty) ->
          printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);

  }

| SIMPL term DOT {

    match simpl_term $2 state.def state.coercion state.oracles state.overload state.implicite with
      | None -> printf "Cannot Simplify\n\n";
      | Some (te, ty) ->
          printf "%s: %s\n\n" (string_of_term te state.implicite) (string_of_term ty state.implicite);

  }

| SHOWDEF DOT {

    let l = vmext state.def in
      
      printf "Definitions:\n";

      List.fold_left (

        fun tl hd ->
          let (name, (te, ty)) = hd in
            printf "%s := %s : %s\n"
              name
              (string_of_term te state.implicite)
              (string_of_term ty state.implicite);

      ) () l; printf "\n";

  }

| SHOWCOERCION DOT {

    let l = state.coercion in
      
      printf "Coercions:\n";

      List.fold_left (

        fun tl hd ->
          let (te, ty, i) = hd in
            printf " %s: %s  [%d]\n"
              (string_of_term te state.implicite)
              (string_of_term ty state.implicite)
              i;

      ) () l; printf "\n";

  }

| QUIT DOT {
    exit 0;
  }

;

  /**********************************************/
  /* Hommage a ceux qui ont montrer la voie */

    funclambda:

| listtermdef term DOT {

    let (lname, l) = List.split $1 in
    let (lte, lty) = List.split l in
      
    match check_mutualterm lname lte lty state.def state.coercion state.oracles state.overload state.implicite with
      | None -> None
      | Some l ->
          
          let def = 
            
            List.fold_left (
            
              fun acc hd ->
                
                let (name, (te, ty)) = hd in
                  (*
                  printf "%s := %s : %s\n"
                    name
                    (string_of_term te state.implicite)
                    (string_of_term ty state.implicite);
                  *)
                  let te = var2cste te (list2varset lname) in
                  let ty = var2cste ty (list2varset lname) in
                    VarMap.add name (te, ty) acc 

            ) VarMap.empty l in
            
            match check_term $2 None (varmap_union def state.def) state.coercion state.oracles state.overload state.implicite true with
              | None -> None
              | Some (te, ty) -> 
                  Some (def, te, ty)

  }

| term DOT {

    match check_term $1 None state.def state.coercion state.oracles state.overload state.implicite false with
      | None -> None
      | Some (te, ty) ->
          Some (VarMap.empty, te, ty)

  }

;



  listtermdef: 

| termdef { $1::[] }

| termdef listtermdef { $1 :: $2 }

;

  termdef: 

| LET NAME listargs DDOT term DDOT EQ term IN {

    
    let te = Phi ($2, $3, $5, [], $8) in 
      ($2, (te, None))
  }

| LET NAME listargs DDOT EQ term IN {

    let te = Phi ($2, $3, Var "?T", [], $6) in 
      ($2, (te, None))

  }


;



 term:
| terma {

    $1
      
  }

| terma suiteterm {
    let te1 = $1 in
    let te2 = $2 in
      App (te1,te2)
  }
| term IMPL term {
    
    let te = $1 in
    let te2 = $3 in
    let x = "?H" in
      Forall (x, te, te2)

}
 
;

  terma:

| UNDERSCORE {

    Var "?T"

  }

| NAME { 

    try (
      
      let i = VarMap.find $1 state.implicite in

      let l = list_of_n_elt "?T" i in

        App (Cste $1, (list_name2list_var l))

    ) with
      | _ ->  Var $1

  }

| AROBAS NAME {

    Cste $2

  }

| LAMBDA LPAREN NAME DDOT term RPAREN DOT term {

    let ty = $5 in
    let te = $8 in
      Lambda ($3, ty, te)

  }

| LAMBDA NAME DOT term {

    let te = $4 in
    let x = "?T" in
      Lambda ($2, (Var x), te)

  }

| FORALL LPAREN NAME DDOT term RPAREN DOT term {

    let ty = $5 in
    let te = $8 in
      Forall ($3, ty, te)

  }

| FORALL NAME DOT term {

    let te = $4 in
    let x = "?T" in
      Forall ($2, (Var x), te)

  }

| MATCH term WITH destructors {
    Match ($2, $4, None)
  }

| MATCH term RETURN term WITH destructors {
    Match ($2, $6, Some $4)
  }

| CONS NUM OF term {
    Cons ((int_of_string $2), $4)
  }

| TYPE { Type }

| LPAREN term RPAREN { $2 }

/* Overloaded  */
| term EQ term { App(Cste "eq", $1::$3::[])}
| term PLUS term { App(Cste "plus", $1::$3::[])}
| term MINUS term { App(Cste "minus", $1::$3::[])}
| term MULT term { App(Cste "mult", $1::$3::[])}
| term DIV term { App(Cste "div", $1::$3::[])}


/* Boolean */

| BOOLT { boolType }
| BOOL { create_boolTerm $1 }
| BNEG term { App (bnegTerm, $2::[]) }
| term BAND term { App (bandTerm, $1::$3::[]) }
| term BOR term { App (borTerm, $1::$3::[]) }

/* Num */
| NUM { create_numTerm (num_of_string $1) }

/* Only for testing */


| PHI LPAREN NAME COMMA LPAREN context RPAREN COMMA term COMMA LBRACK listnum RBRACK RPAREN DOT term {
    Phi ($3, $6, $9, $12, $16)
  }

| IND LPAREN NAME COMMA LPAREN context RPAREN COMMA term RPAREN DOT listterm {
    Ind ($3, $6, $9, $12)
  }


;

  listterm:
| term { $1 :: [] }
| listterm COMMA listterm { $1 @ $3 }
| LPAREN listterm RPAREN { $2 }
| LPAREN RPAREN { [] }
;

  suiteterm:
| terma { $1::[] }
| terma suiteterm { $1 :: $2 }
;


  args:

| LPAREN NAME DDOT term RPAREN { ($2,$4) }
| NAME { 
        
    ($1, Var "?T")

  }
;

  context:
| args { $1::[] }
| args COMMA context { $1::$3 }
| { [] }
;

  listargs:

| args { $1::[] }
| args listargs { $1 :: $2}
| { [] }
;

 destructors:
| destructors destructors {$1 @ $2 }
| PAR term { $2::[] }
| { [] } 
;

 listnum:
| NUM { (int_of_string $1)::[] }
| listnum COMMA listnum {$1 @ $3}
| { [] }
;

  constructors:
| constructors constructors { $1 @ $2 }
| PAR NAME DDOT term { ($2, $4)::[] }
| { [] }
;

